/*
 * Minimum supported browser: Chrome 32
 */


var commandName = null;
var command = null;

var c_ApplicationName = 'net.spoon.chromenativehost';

var c_ErrorInvalidRequest = 'Invalid request';
var c_ErrorInvalidCommand = 'Invalid command';
var c_ErrorInvalidResponse = 'Invalid response from native host';
var c_ErrorCommandFailed = 'Failed to execute command';

// some constants used for responses from extension tasks
var Response = {
  ok: function(id) {
    return {
      type: "FROM_SPOON_EXT",
      id: id,
      response: {
        succeeded: true
      }
    };
  },
  error: function(id, msg) {
    return {
      type: "FROM_SPOON_EXT",
      id: id,
      response: {
        succeeded: false,
        msg: msg
      }
    };
  }
};

// used for communication between extension and sandman over websocket
function WebSocketTask(id, port, command) {
  var c_MaxTimeout = 10000;
  var c_WebSocketWaitTime = 500;

  function _Start() {
    return new Promise(function (resolve, reject) {
      var m_startTime = new Date().getTime();

      if(port === 0) {
        console.error('Invalid port: %s', port);
        var msg = Response.error(id, 'Invalid port');
        reject(msg);
        return;
      }

      var m_pending = true;
      var m_ws = new WebSocket('ws://127.0.0.1:' + port + '/sandman');

      m_ws.onopen = function () {
        m_ws.send(command);
      };

      m_ws.onmessage = function (resp) {
        if (m_pending) {
          m_pending = false;

          if (resp.data === 'true') {
            resolve(Response.ok(id));
          } else if (resp.data === 'false') {
            resolve(Response.error(id, "response was false"));
          } else if (!resp.data) {
            resolve(Response.error(id, "data not defined"));
          } else if (typeof resp.data === 'string' && resp.data.length == 0) {
            resolve(Response.error(id, "empty response"));
          } else {
            var msg = Response.ok(id);
            msg.response.result = resp.data;
            resolve(msg);
          }
        }

        m_ws.close();
      };

      m_ws.onerror = function (resp) {
        console.debug("Error received on websocket request");

        if (m_pending) {
          m_pending = false;

          var msg = Response.error(id);
          msg.response.result = resp.data;
          reject(msg);
        }

        m_ws.close();
      };

      // wait for response
      function _Wait() {
        var m_currentTime = new Date().getTime();

        if (!m_pending) {
          return;
        }

        if (m_currentTime - m_startTime < c_MaxTimeout) {
          setTimeout(_Wait, c_WebSocketWaitTime);
        } else {
          console.debug("Timeout waiting for command to complete");
          m_pending = false;
          reject(Response.error(id, "Timeout"));
        }
      }

      _Wait();
    });
  }

  return {
    Start: _Start
  };
}

// object used to send messages to the native host process
function GenericNativeCommand(commmand) {
  var m_pending = false;
  var m_succeeded = false;
  var m_result = null;
  var that = {};

  that.Start = function(id, sendResponse) {
    function _OnError(msg) {
      m_succeeded = false;
      sendResponse(Response.error(id, msg));
    }

    function _NativeCallback(response) {
      var msg = null;

      m_pending = false;

      if (typeof(response) == "undefined") {
        if(typeof(chrome.runtime.lastError) !== 'undefined') {
          console.error(c_ErrorCommandFailed, chrome.runtime.lastError.message);
        }

        _OnError(c_ErrorCommandFailed);
        return;
      }

      if (!response.hasOwnProperty('code')) {
        console.error(c_ErrorCommandFailed, c_ErrorInvalidResponse, response);

        _OnError(c_ErrorCommandFailed);
        return;
      }

      if (response.code !== 'ok' || !commmand.ValidateResponse(response)) {
        console.error(c_ErrorCommandFailed, c_ErrorInvalidResponse, response);
        _OnError(c_ErrorCommandFailed);
        return;
      }

      m_succeeded = true;
      m_result = commmand.ProcessResponse(id, response);

      sendResponse(m_result);
    }

    m_pending = true;
    m_succeeded = false;
    chrome.runtime.sendNativeMessage(
      c_ApplicationName, {
        command: commmand.Name
      },
      _NativeCallback);
  };

  return that;
}

// command to restart the sandman
function RestartCommand() {
  var restartCmd = {
    Name: 'restart',
    ValidateResponse: function(resp) {
      return true;
    },
    ProcessResponse: function(id, resp) {
      return Response.ok(id);
    }
  };

  return GenericNativeCommand(restartCmd);
}

// command to get the version of the sandman
function VersionCommand() {
  var versionCmd = {
    Name: 'version',
    ValidateResponse: function(resp) {
      return typeof resp.version === 'string';
    },
    ProcessResponse: function(id, resp) {
      var result = Response.ok(id);
      result.response.version = resp.version;
      return result;
    }
  };

  return GenericNativeCommand(versionCmd);
}

// command to get the current status of the sandman (including port where websocket server is listening)
function StatusCommand() {
  var statusCmd = {
    Name: 'status',
    ValidateResponse: function(resp) {
      // ignore type checking for trackingId for backward compatibility
      return typeof resp.port === 'number' && typeof resp.identity === 'string'
    },
    ProcessResponse: function(id, resp) {
      var result = Response.ok(id);
      result.response.port = resp.port;
      result.response.identity = resp.identity;
      
      if(typeof (resp.trackingId) === 'string' && resp.trackingId.length > 0) {
        // use trackingId assigned by sandbox manager
        // if not possible (browser launched manually or extension connected to legacy message host), trackingId will be generated by extension
        trackingId.set(resp.trackingId);
      }
      
      result.response.trackingId = trackingId.get();
      
      return result;
    }
  };

  return GenericNativeCommand(statusCmd);
}

var trackingId = function () {
    var id = null;

    function generateUUID() {
        var d = new Date().getTime();
        var uuid = 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function (c) {
            var r = (d + Math.random() * 16) % 16 | 0;
            d = Math.floor(d / 16);
            return (c == 'x' ? r : (r & 0x3 | 0x8)).toString(16);
        });
        return uuid;
    }

    return {
        get: function () {
            if (id === null) {
                id = generateUUID();
            }
            return id;
        },
        set: function (value) { id = value; }
    };
}();

// sandman connection info
var sandmanConnection = function() {
    var that = {
        version: null,
        port: 0,
        identity: null,
        trackingId: null,
        initializationTask: null,
        failures: 0
    };

  function _reset() {
    that.port = 0;
    that.identity = null;
    that.trackingId = null;
  }

  function _set(port, identity, trackingId) {
    that.port = port;
    that.identity = identity;
    that.trackingId = trackingId;
    
    console.info(
      'Initialized sanbox manager: port %s, identity: %s, trackingId: \'%s\'',
      that.port,
      that.identity,
      that.trackingId);
  }

  function _isInitialized() {
    return that.port !== 0 && that.identity !== null && that.trackingId !== null;
  }

  function _succeeded(msg) {
    return msg.response.succeeded;
  }

  function _rawGetStatus(successCallback, errorCallback) {
    StatusCommand().Start(c_ApplicationName, function(msg) {
      if(_succeeded(msg)) {
        that.failures = 0;

        successCallback(msg);
      } else {
        that.failures += 1;

        errorCallback(msg);
      }
    });
  }

  function _refreshVersion() {
      VersionCommand().Start(0,
          function (msg) {
              if (_succeeded(msg)) {
                  that.version = msg.response.version;
                  console.log("Client version " + that.version);
              }
              else {
                  console.error("Unable to get client version");
                  that.version = null;
              }
          });
  }

  function _isVersionGreaterThanOrEqual(version) {
      if (that.version == null)
          return false;

      var partsClient = that.version.split('.');
      var partsReq = version.split('.');

      while (partsClient.length < partsReq.length) partsClient.push("0");
      while (partsReq.length < partsClient.length) partsReq.push("0");
      
      for (var i = 0; i < partsClient.length; ++i) {
        var req = Number(partsReq[i]);
        var cli = Number(partsClient[i]);
        if (req > cli)
          return false;
        else if(req < cli)
          return true;
      }

      return true;
  }

  function _rawPing(successCallback, errorCallback) {
    var command = [that.identity, 'ping'].join(';');
    // ignore request id coming from background script
    var task = WebSocketTask('dummy_id', that.port, command);
    task.Start().then(successCallback, errorCallback);
  }

  function _rawRestart(successCallback, errorCallback) {
    RestartCommand().Start(c_ApplicationName, function(msg) {
      if(_succeeded(msg)) {
        that.failures = 0;

        successCallback(msg);
      } else {
        that.failures += 1;

        errorCallback(msg);
      }
    });
  }

  function _restartWithGetStatus(successCallback, errorCallback) {
    _reset();

    var onRestart = function(msg) {
      var onStatus = function(msg) {
        var response = msg.response;
        _set(response.port, response.identity, response.trackingId);

        successCallback(msg);
      };

      _rawGetStatus(onStatus, errorCallback);
    };

    var onError = function(err) {
      console.error('Failed to restart sandbox mananger', err);

      errorCallback(err);
    };

    _rawRestart(onRestart, onError);
  }

  function _pingWithRestartOnFailure(successCallback, errorCallback) {
    _rawPing(successCallback, function() { _restartWithGetStatus(successCallback, errorCallback) } );
  }

  function _doInitialize() {
    return new Promise(function(resolve, reject) {
      function _doResolve(res) {
        that.initializationTask = null;
        resolve(res);
      }

      function _doReject(res) {
        that.initializationTask = null;
        _reset();
        reject(res);
      }

      if(_isInitialized()) {
        _pingWithRestartOnFailure(_doResolve, _doReject);
      } else {
        var onGetStatus = function(msg) {
          var response = msg.response;
          _set(response.port, response.identity, response.trackingId);
          _pingWithRestartOnFailure(_doResolve, _doReject);
        };
        _rawGetStatus(onGetStatus, _doReject);
        _refreshVersion();
      }
    });
  }

  function _doRestart() {
    return new Promise(function(resolve, reject) {
      function _doResolve(res) {
        that.initializationTask = null;
        resolve(res);
      }

      function _doReject(res) {
        that.initializationTask = null;
        _reset();
        reject(res);
      }

      _restartWithGetStatus(_doResolve, _doReject);
    });
  }

  function _initialize(successFunc, failedFunc) {
    if(that.initializationTask === null) {
      that.initializationTask = _doInitialize();
    }

    if (typeof (successFunc) === 'function') {
        if (typeof (failedFunc) !== 'function')
            failedFunc = successFunc;
        that.initializationTask.then(successFunc, failedFunc);
    }
  }

  function _restart(callback) {
    if(that.initializationTask === null) {
      that.initializationTask = _doRestart();
    }

    if(typeof (callback) !== 'undefined') {
      that.initializationTask.then(callback, callback);
    }
  }

  that.isInitialized = _isInitialized;
  that.initialize = _initialize;
  that.restart = _restart;
  that.isVersionGreaterThanOrEqual = _isVersionGreaterThanOrEqual;

  return that;
}();

var navigationTrigger = function() {
    var resumedNavigations = {};

    function _removeRequest(tabId) {
      delete resumedNavigations[tabId];
    }

    return {
        isResumed: function(request) {
            return (request.tabId in resumedNavigations) && resumedNavigations[request.tabId].url === request.url;
        },
        resume: function(request) {
            resumedNavigations[request.tabId] = request;
            chrome.tabs.update(request.tabId, {url: request.url}, function(tab) {
                console.debug('Resumed navigation to %s on tab %s', tab.url, tab.id);
            });
        },
        complete: function(request) {
            _removeRequest(request.tabId);
        },
        install: function() {
            chrome.tabs.onRemoved.addListener(_removeRequest);
        }
    };
}();

var beforeRequestHandler = function() {
    var that = {};
    var openedTabs = new Set();
    var redirectorPatterns = null;

    function _isClientSupported(version) {
        if (sandmanConnection.version === null)
            return true;
        return sandmanConnection.isVersionGreaterThanOrEqual(version);
    }

    var handler = function (request) {
    
        if (redirectorPatterns !== null && redirectorPatterns.length == 0)
          return; // no rules so don't process anything

        if(sandmanConnection.failures > 16) {
            console.warn('beforeRequestHandler failed to connect to sandbox manager and message host %d times in a row. Deactivating Turbo.net Browser Helper.',
                sandmanConnection.failures
            );
            that.uninstall();
            return;
        }

        if(typeof request !== 'object'
            || !request.hasOwnProperty('method')
            || !request.hasOwnProperty('type')
            || !request.hasOwnProperty('url')
            || !request.hasOwnProperty('tabId')) {
            console.error('Error: beforeRequestHandler failed to process request: %s', request);
            return;
        }

        if(request.method !== 'GET') {
            return;
        }

        if(request.type !== 'main_frame') {
            // ignore requests for 'sub_frame', font', 'image', 'other', 'object', 'script', 'stylesheet', 'xmlhttprequest'
            return;
        }

        // check if for chrome (could also be firefox and maybe others in the future). this stopped being a valid check in chrome61.
        var match = navigator.userAgent.match(/Chrom(e|ium)\/([0-9]+)\./i);
        if(match !== null && parseInt(match[2], 10) < 61 && !openedTabs.has(request.tabId)) {
            // ignore requests triggered by Chrome prediction service
            // Chrome prediction service requests are GET requests called from a background tab
            // after navigation is confirmed the source tab is closed and replaced by the background tab used by the Chrome prediction service
            // it happens without calling webRequest lifecycle methods
            // cancel Chrome prediction service requests
            
            // todo: this is kind of bad since it will reduce perf. anything else we can do about this?

            console.debug('Cancel request for %s on tab %s', request.url, request.tabId);
            return {redirectUrl: 'javascript:'};
        }

        if(navigationTrigger.isResumed(request)) {
            navigationTrigger.complete(request);
            return;
        }

        var onBrowseCompleted = function(msg) {
            var response = msg.response;
            var tabId = request.tabId;
            var url = request.url;

            if (!response.succeeded) {
                console.error('Failed to process browse response', msg);
                
                console.info('Legacy sandbox manager does not support redirector. The feature will be deactivated.');
                that.uninstall();
                navigationTrigger.resume(request);
                
                return;
            }

            if (response.result === 'redirect;') {
                console.debug('Redirection triggerd for %s on tab %s', url, tabId);

                chrome.tabs.sendMessage(tabId, {type: 'is_dom_empty'}, function(response) {
                    // response is undefined if content script is not loaded
                    // case of about:blank and pages when natigation was cancelled on the first accessed url
                    var isPageEmpty = typeof response === 'undefined' || response.result;
                    if(isPageEmpty) {
                        // close empty tab
                        chrome.tabs.get(tabId, function(tab) {
                            if (typeof(tab) !== 'undefined') {
                                chrome.tabs.remove(tabId);
                            }
                        });
                    } else {
                        // show redirection message
                        chrome.tabs.sendMessage(tabId, {type: 'insert_redirect_msg'});
                    }
                });
            } else {
                navigationTrigger.resume(request);
            }
        };

        function startBrowseQuery() {
            // trackingId is used for break redirection loop algorithm to work
            // see client\Manager\Redirector\BrowserRedirector.cs for more info
            var command = [sandmanConnection.identity, 'browse', request.url, sandmanConnection.trackingId].join(';');
            var query = WebSocketTask(request.requestId, sandmanConnection.port, command);

            console.debug('Redirector query for %s on tab %s', request.url, request.tabId);

            return query.Start();
        }

        function browseQueryNoRetry() {
            if(!_isClientSupported("3.33.1066")) {
                console.log("Client doesn't support browse command.");
                navigationTrigger.resume(request);
            }
            else {
                startBrowseQuery().then(onBrowseCompleted, function(err) {
                    console.error('Failed to execute browse query. Ignoring Turbo.net Browser Helper configuration.');

                    navigationTrigger.resume(request);
                });
            }
        }

        function browseQueryRetry() {
            if (!_isClientSupported("3.33.1066")) {
                console.log("Client doesn't support browse command.");
                navigationTrigger.resume(request);
            }
            else {
                startBrowseQuery().then(onBrowseCompleted, function (err) {
                    console.error('Failed to execute browse query. Restarting sandbox manager to retry.');

                    sandmanConnection.restart(browseQueryNoRetry);
                });
            }
        }

        // check if there is a matching pattern
        var found = false;
        if (redirectorPatterns !== null) {
            for (var i = 0; i < redirectorPatterns.length; ++i) {
                if (redirectorPatterns[i].test(request.url)) {
                    found = true;

                    break;
                }
            }
        }

        if (found || redirectorPatterns === null) {
            // send off to the sandman to deal with
            if (sandmanConnection.isInitialized())
                browseQueryRetry();
            else
                sandmanConnection.initialize(browseQueryNoRetry);

            return { redirectUrl: 'javascript:' };
        }

    };

    function startGetRulesQuery() {
        var command = [sandmanConnection.identity, 'getrules'].join(';');
        var query = WebSocketTask(0, sandmanConnection.port, command);

        console.debug('Redirector query for new rules');

        return query.Start();
    }

    function getRulesMessageHandler(msg) {
        var res = msg.response;
        if (res.succeeded) {
            // parse results
            var p = [];
            var obj = JSON.parse(res.result);
            for (var i = 0; i < obj.patterns.length; ++i) {
                p.push(new RegExp(obj.patterns[i], "i"));
            }

            redirectorPatterns = p;
        }
    }

    function getRulesQueryNoRetry(refreshTimeout, errFunc) {
        if (!_isClientSupported("3.33.1070")) {
            console.log("Client doesn't support getrules command.");
            redirectorPatterns = []; // client doesn't support this message so default to no rules
        }
        else {
            startGetRulesQuery().then(
                function (msg) {
                    getRulesMessageHandler(msg);

                    if (refreshTimeout)
                        setTimeout(refreshRules, refreshTimeout, refreshTimeout);
                },
                function (err) {
                    console.error('Failed to execute getrules query');
                    if (typeof (errFunc) == "function")
                        errFunc(err);
                    else
                        redirectorPatterns = []; // client doesn't support this message so default to no rules
                });
        }
    }

    function getRulesQueryRetry(refreshTimeout) {
        getRulesQueryNoRetry(
            refreshTimeout,
            function (err) {
                console.error('Restarting sandbox manager to retry.');
                sandmanConnection.restart(function () { getRulesQueryNoRetry(refreshTimeout) });
            });
    }

    function refreshRules(refreshTimeout) {
        if (sandmanConnection.isInitialized())
            getRulesQueryRetry(refreshTimeout);
        else
            sandmanConnection.initialize(
                function () {
                    getRulesQueryNoRetry(refreshTimeout)
                },
                function () {
                    // can't initialize the sandman so set empty rule list so we don't go to the sandman ever (CLNT-6202)
                    redirectorPatterns = [];
                });
    }

    refreshRules(60000); // refresh the rules patterns every minute
  
    var onTabCreated = function(tab) {
        openedTabs.add(tab.id);
        refreshRules();
    };
  
    var onTabRemoved = function(tabId) {
        openedTabs.delete(tabId);
    };

    that.uninstall = function() {
        chrome.tabs.onRemoved.removeListener(onTabRemoved);
        chrome.tabs.onCreated.removeListener(onTabCreated);
        chrome.webRequest.onBeforeRequest.removeListener(handler);
        
        openedTabs.clear();
    };

    that.install = function() {
        chrome.tabs.query({}, function(results) {
            var length = results.length;
            for(var index = 0; index < length; ++index) {
                openedTabs.add(results[index].id);    
            }
        });
        
        chrome.tabs.onRemoved.addListener(onTabRemoved);
        chrome.tabs.onCreated.addListener(onTabCreated);
        chrome.webRequest.onBeforeRequest.addListener(handler, { urls: ['<all_urls>'] }, ['blocking']);
    }

    return that;
}();

var connectionHandler = function() {
  var that = {};
  
  that._handler = function (port) {
    console.debug("Connection received");

    port.onMessage.addListener(
      // handler for when a message from the page scripts is received
      function (request) {
        var command = null;

        if (typeof request !== 'object' || !request.hasOwnProperty("id")) {
          port.postMessage(Response.error(-1, c_ErrorInvalidRequest));
          return;
        }

        if (!request.hasOwnProperty('command')) {
          port.postMessage(Response.error(request.id, c_ErrorInvalidRequest));
          return;
        }

        function _PostResponse(msg) {
          port.postMessage(msg);
        }

        function _Succeeded(msg) {
          return typeof (msg) !== 'undefined' && msg.hasOwnProperty('response') && msg.response.succeeded;
        }

        commandName = request.command.toLocaleLowerCase();
        switch (commandName) {
          case 'restart':
            sandmanConnection.restart(function(msg) {
              var msgToUse = null;
              if(msg.response.succeeded) {
                msgToUse = Response.ok(request.id);
              } else {
                msgToUse = Response.error(request.id);
              }
              _PostResponse(msgToUse);
            });
            break;

          case 'status':
            StatusCommand().Start(request.id, _PostResponse);
            break;

          case 'version':
            VersionCommand().Start(request.id, _PostResponse);
            break;

          case 'websocket':
            {
              if (!request.hasOwnProperty("params") ||
                !request.params.hasOwnProperty("port") ||
                !request.params.hasOwnProperty("command")) {
                port.postMessage(Response.error(request.id, c_ErrorInvalidRequest));
                return;
              }

              console.log(request.id + ": " + request.params.command);

              var task = WebSocketTask(request.id, request.params.port, request.params.command);
              task.Start().then(_PostResponse, _PostResponse);
            }
            break;

          default:
            port.postMessage(Response.error(request.id, c_ErrorInvalidCommand));
            return;
        }
      });
  };
  
  that.install = function() {
    chrome.runtime.onConnect.addListener(that._handler);
  }
  
  return that;
}();

var contentScript = function() {
  function __insertContentScriptForTab(tab) {
    if (tab.status !== 'complete') {
      return;
    }
    
    if(tab.url.startsWith('chrome://') && tab.url !== 'chrome://newtab/') {
      // do not insert content scripts on native browser pages like chrome://extensions/
      return;
    }
    
    chrome.tabs.executeScript(tab.id, { file: 'content.js', allFrames: true });
    chrome.tabs.executeScript(tab.id, { file: 'browserhelper.js', allFrames: false });
  }

  function __insertContentScriptForTabs(tabs) {
    var length = tabs.length;
    for(var i = 0; i < length; ++i) {
        __insertContentScriptForTab(tabs[i]);
    }
  }

  return {
    install: function() {
      chrome.tabs.onActivated.addListener(
        function(Id, changeInfo, tab) {
          chrome.tabs.get(Id.tabId, __insertContentScriptForTab);
        });
      chrome.tabs.onUpdated.addListener(
        function(Id, changeInfo, tab) {
          __insertContentScriptForTab(tab);
        });
      
      // do not insert script on panel, app and devtools window
      chrome.tabs.query({ active: true, windowType: 'normal' }, __insertContentScriptForTabs);
      chrome.tabs.query({ active: true, windowType: 'popup' }, __insertContentScriptForTabs);
    }
  };
}();

// listen for tabs
navigationTrigger.install();

// listen for connections
connectionHandler.install();

// listen for web requests
beforeRequestHandler.install();

// insert content script
contentScript.install();
